#include "plumpconfig.h"
#include <linux/ptrace.h>

int ptrace_okay()
{
#ifdef MEDIUM_HAS_PTRACE
	#ifdef OS_SUNOS
		return PLUMP_ERROR_NOTSUPPORTED;
	#endif
	return PLUMP_SUCCESS;
#else
	return PLUMP_ERROR_NOTSUPPORTED;
#endif
}

#ifdef MEDIUM_HAS_PTRACE

int ptrace_execContinueSingleStep(plump_t *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags);
int ptrace_execContinueCont(plump_t *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags);

#include <sys/wait.h>

#if defined(OS_LINUX)
	#define PTRACE_ADDR_CAST  void *
	#define REGISTERS_TYPE    REGISTER_STRUCT_TYPE_USER_REGS_STRUCT
	#define REGISTERS         struct user_regs_struct
#elif defined(OS_FREEBSD)
	#define PTRACE_ATTACH     PT_ATTACH
	#define PTRACE_DETACH     PT_DETACH

	#define PTRACE_PEEKDATA   PT_READ_D
	#define PTRACE_POKEDATA   PT_WRITE_D

	#define PTRACE_SETREGS    PT_SETREGS
	#define PTRACE_GETREGS    PT_GETREGS
 
	#define PTRACE_CONT       PT_CONTINUE

	#define PTRACE_SINGLESTEP PT_STEP

	#define PTRACE_ADDR_CAST  void *
	#define REGISTERS_TYPE    REGISTER_STRUCT_TYPE_REG
	#define REGISTERS         struct reg
#else
	#include <sys/ptrace.h>
#endif

plumpmedium_t ptrace_getMedium()
{
	return PLUMP_MEDIUM_PTRACE;
}

int ptrace_open(plump_t *p, int pid, unsigned long access)
{
	int ret = PLUMP_ERROR_NOTSUPPORTED;

	if (ptrace(PTRACE_ATTACH, pid, 0, 0) < 0)
	{
		switch (errno)
		{
			case ESRCH:
				ret = PLUMP_ERROR_NOSUCHPID;
				break;
			default:
				ret = PLUMP_ERROR_FAILURE;
				break;
		}
	}
	else
	{
		// Wait for ptrace to attach and stop the process.
		wait(NULL);

		p->state |= PLUMP_STATE_STOPPED;

		ret = PLUMP_SUCCESS;
	}

	return ret;
}

int ptrace_close(plump_t *p)
{
	int ret = PLUMP_ERROR_NOTSUPPORTED;

	// If the process is in a stopped state, continue it indefinitely.
	if (p->state & PLUMP_STATE_STOPPED)
		ptrace_execContinue(p, 0, NULL, PLUMP_CONTINUE_FLAG_INDEFINITELY);

	// syscall PTRACE_DETACH
#if defined(OS_LINUX)
	if (ptrace(PTRACE_DETACH, p->pid, 0, 0) < 0)
#else
	if (ptrace(PTRACE_DETACH, p->pid, (PTRACE_ADDR_CAST)1, 0) < 0)
#endif
	{
		switch (errno)
		{
			case ESRCH:
				ret = PLUMP_ERROR_NOSUCHPID;
				break;
			default:
				ret = PLUMP_ERROR_FAILURE;
				break;
		}
	}
	else
		ret = PLUMP_SUCCESS;

	return ret;
}

int ptrace_execContinue(plump_t *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags)
{
	int ret     = PLUMP_ERROR_NOTSUPPORTED;
	int request = 0;
	

	// Are we stopped?
	if (!(p->state & PLUMP_STATE_STOPPED))
		return PLUMP_ERROR_FAILURE;

	do
	{
		/*
		 * a) If the user wants to single step then we do so.
		 * b) If they want us to run until we hit a user break point, then we need to single step.
		 * c) If they want us to run until we hit a stop address, then we single step.
		 */
		if ((flags & PLUMP_CONTINUE_FLAG_SINGLESTEP) ||
				(flags & PLUMP_CONTINUE_FLAG_UNTILUSRBP) ||
				(stop && *stop != 0))
			request = PTRACE_SINGLESTEP;
		else // Otherwise, we continue.
			request = PTRACE_CONT;

#if defined(OS_LINUX) || defined(OS_FREEBSD)
		/*
		 * In Linux/FreeBSD we need to update the program counter to the start address
		 * if it's non-zero.
		 */
#if defined(OS_LINUX)
		if (start)
#elif defined(OS_FREEBSD)  // If we singlestep in freebsd we need to execution divert the same way we do in Linux.
		if ((start) && 
				(request == PTRACE_SINGLESTEP))
#endif
		{
			plumpreg_t pc[1];
	
			pc[0].name    = REGISTER_TYPE_PROGRAM_COUNTER;
			pc[0].val.b32 = start;
	
			// If we fail to set the pc register, break out.
			if ((ret = ptrace_registersSet(p, pc, 1)) != PLUMP_SUCCESS)
				break;
		}
#endif

		// Depending on the type of request, we call a different function.
		switch (request)
		{
			case PTRACE_SINGLESTEP:
				ret = ptrace_execContinueSingleStep(p, start, stop, flags);
				break;
			case PTRACE_CONT:
				ret = ptrace_execContinueCont(p, start, stop, flags);
				break;
			default:
				break;
		}

		// If we were successfully and we aren't running indefinitely AND stop is valid, we grab the stop addr.
		if ((ret >= PLUMP_SUCCESS) && (flags) && (!(flags & PLUMP_CONTINUE_FLAG_INDEFINITELY)) && (stop))
		{
			plumpreg_t pc[1];
	
			pc[0].name    = REGISTER_TYPE_PROGRAM_COUNTER;
			pc[0].val.b32 = 0;

			if ((ret = ptrace_registersGet(p, pc, 1)) != PLUMP_SUCCESS)
				break;

			*stop = pc[0].val.b32;
		}

	} while (0);

	return ret;
}

int ptrace_execContinueSingleStep(plump_t *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags)
{
	int ret = PLUMP_ERROR_FAILURE;
	int ptret;
	plumpreg_t pc[1];

	pc[0].name    = REGISTER_TYPE_PROGRAM_COUNTER;
	pc[0].val.b32 = 0;

	// If we aren't stopped, do stop.
	if (!(p->state & PLUMP_STATE_STOPPED))
		ptrace_execStop(p);

#if defined(OS_LINUX)
	while ((ptret = ptrace(PTRACE_SINGLESTEP, p->pid, 0, 0)) >= 0)
#elif defined(OS_FREEBSD)
	while ((ptret = ptrace(PTRACE_SINGLESTEP, p->pid, (PTRACE_ADDR_CAST)1, 0)) >= 0)
#endif
	{
		// Grab out SIGTRAP
		wait(NULL);

		// If we're merely single stepping, break out the first iteration
		if (flags & PLUMP_CONTINUE_FLAG_SINGLESTEP)
			break;

		// Get the program counter register
		if ((ret = ptrace_registersGet(p, pc, 1)) != PLUMP_SUCCESS)
			break;

		// If we're watching for a stop address
		if (stop && *stop != 0)
		{
			// Is this it?
			if (*stop == pc[0].val.b32)
				break;
		}

		// If we're watching for user breakpoints
		if ((flags & PLUMP_CONTINUE_FLAG_UNTILUSRBP) && 
				(plump_breakpointIsWatching(p, pc[0].val.b32) == PLUMP_SUCCESS))
			break;
	}
	
	if (ptret < 0)
	{
		switch (errno)
		{
			case ESRCH:
				ret = PLUMP_ERROR_NOSUCHPID;
				break;
			default:
				ret = PLUMP_ERROR_FAILURE;
				break;
		}
	}
	else
		ret = PLUMP_SUCCESS;

	return ret;
}

int ptrace_execContinueCont(plump_t *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags)
{
	int ret = PLUMP_ERROR_FAILURE;

	do
	{
#if defined(OS_LINUX)
		// If the start address is non-zero, we divert via changing the program counter
		if (ptrace(PTRACE_CONT, p->pid, 0, 0) < 0)
#elif defined(OS_FREEBSD)
		if (ptrace(PTRACE_CONT, p->pid, (PTRACE_ADDR_CAST)((start == 0) ? 1 : start), 0) < 0)
#endif
		{
			switch (errno)
			{
				case ESRCH:
					ret = PLUMP_ERROR_NOSUCHPID;
					break;
				default:
					ret = PLUMP_ERROR_FAILURE;
					break;
			}

			break;
		}
			
		ret       = PLUMP_SUCCESS;
		p->state &= ~(PLUMP_STATE_STOPPED);

		// If this is an indefinite continue, break out.
		if ((!flags) || 
				(flags & PLUMP_CONTINUE_FLAG_INDEFINITELY))
			break;

		if (flags & PLUMP_CONTINUE_FLAG_UNTILSRCBP)
		{
			int        status;

			while (wait(&status) > 0)
			{
				if (!WIFSTOPPED(status))
					continue;
			
				p->state |= PLUMP_STATE_STOPPED;

				// Is this a source breakpoint?
				if (WSTOPSIG(status) == SIGTRAP)
				{
					ret = PLUMP_SUCCESS_SRCBREAKPOINTHIT;

					break;
				}

				p->state &= ~(PLUMP_STATE_STOPPED);

				// Next please.
#if defined(OS_LINUX)
				ptrace(PTRACE_CONT, p->pid, 0, 0);
#elif defined(OS_FREEBSD)
				ptrace(PTRACE_CONT, p->pid, (PTRACE_ADDR_CAST)1, 0);
#endif
			}
		}

	} while (0);

	return ret;
}

int ptrace_execStop(plump_t *p)
{
	int ret = PLUMP_ERROR_NOTSUPPORTED;

	// Are we already stopped?
	if (p->state & PLUMP_STATE_STOPPED)
		return PLUMP_ERROR_ALREADY;

	// Send a SIGSTOP.
	kill(p->pid, SIGSTOP);

	// Wait for our attached child to stop.
	wait(NULL);

	p->state |= PLUMP_STATE_STOPPED;
	ret       = PLUMP_SUCCESS;

	return ret;
}

int ptrace_memoryWrite(plump_t *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength, unsigned long *bytesWritten)
{
	plumpaddr_t curr = addr, end = addr + bufferLength;
	unsigned long data = 0, left = bufferLength, y, poke;
	int ret = PLUMP_SUCCESS;

	errno = 0;

	for (;
			ret == PLUMP_SUCCESS && curr < end; 
			curr += 4)
	{
		data = ptrace(PTRACE_PEEKDATA, p->pid, (PTRACE_ADDR_CAST)curr, 0);

		for (y = 0; y < sizeof(data); y++)
		{
			if (left == 0)
				((unsigned char *)&poke)[y] = ((unsigned char *)&data)[y];
			else
				((unsigned char *)&poke)[y] = ((unsigned char *)buffer)[bufferLength - left--];
		}

		if (ptrace(PTRACE_POKEDATA, p->pid, (PTRACE_ADDR_CAST)curr, poke) < 0)
		{
			switch (errno)
			{
#if defined(OS_LINUX)
				case EIO:
#elif defined(OS_FREEBSD) || defined(OS_SUNOS)
				case EFAULT:
#else
				case 0:
#endif
					if (curr == addr)
						ret = PLUMP_ERROR_NOSUCHADDR;
					break;
				default:
					break;
			}
		}
	}

	if (bytesWritten)
		*bytesWritten = ((curr > end) ? end : curr) - addr;

	return ret;
}

int ptrace_memoryRead(plump_t *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength, unsigned long *bytesRead)
{
	plumpaddr_t curr = addr, end = addr + bufferLength;
	unsigned long data = 0, left = bufferLength, y;
	int ret = PLUMP_SUCCESS;
	
	errno = 0;

	for (;
			ret == PLUMP_SUCCESS && curr < end; 
			curr += 4)
	{
		if ((data = ptrace(PTRACE_PEEKDATA, p->pid, (PTRACE_ADDR_CAST)curr, 0)) == (unsigned long)-1)
		{
			switch (errno)
			{
#if defined(OS_LINUX)
				case EIO:
#elif defined(OS_FREEBSD) || defined(OS_SUNOS)
				case EFAULT:
#else
				case 0:
#endif
					if (curr == addr)
						ret = PLUMP_ERROR_NOSUCHADDR;
					break;
				default:
					break;
			}
		}

		for (y = 0; y < sizeof(data) && left != 0; y++)
			((unsigned char *)buffer)[bufferLength - left--] = ((unsigned char *)&data)[y];
	}

	if (bytesRead)
		*bytesRead = bufferLength - left;

	return ret;
}

int ptrace_memoryIsValid(plump_t *p, plumpaddr_t addr)
{
	unsigned long tmp = 0;

	return ptrace_memoryRead(p, addr, &tmp, sizeof(tmp), NULL);
}

int ptrace_memoryGetEndOfSegment(plump_t *p, plumpaddr_t addr, plumpaddr_t *end)
{
	for (;
			ptrace_memoryIsValid(p, addr) == PLUMP_SUCCESS;
			addr += 4);

	if (end)
		*end = addr;

	return PLUMP_SUCCESS;
}

int ptrace_registersSet(plump_t *p, plumpreg_t *registers, unsigned long numRegisters)
{
	int ret = PLUMP_ERROR_NOTSUPPORTED;
	REGISTERS  reg;

#if defined(OS_LINUX)
	if (ptrace(PTRACE_GETREGS, p->pid, 0, (unsigned long)&reg) < 0)
#else
	if (ptrace(PTRACE_GETREGS, p->pid, (PTRACE_ADDR_CAST)&reg, 0) < 0)
#endif
	{
		switch (errno)
		{
			case ESRCH:
				ret = PLUMP_ERROR_NOSUCHPID;
				break;
			default:
				ret = PLUMP_ERROR_FAILURE;
				break;
		}
	}
	else
	{
		plump_registersTranslate(REGISTERS_TYPE, (void *)&reg, registers, numRegisters, 0);

#if defined(OS_LINUX)
		if (ptrace(PTRACE_SETREGS, p->pid, 0, (unsigned long)&reg) < 0)
#else
		if (ptrace(PTRACE_SETREGS, p->pid, (PTRACE_ADDR_CAST)&reg, 0) < 0)
#endif
			ret = PLUMP_ERROR_FAILURE;
		else
			ret = PLUMP_SUCCESS;
	}


	return ret;
}

int ptrace_registersGet(plump_t *p, plumpreg_t *registers, unsigned long numRegisters)
{
	int ret = PLUMP_ERROR_NOTSUPPORTED;
	REGISTERS reg;

#if defined(OS_LINUX)
	if (ptrace(PTRACE_GETREGS, p->pid, 0, (unsigned long)&reg) < 0)
#else
	if (ptrace(PTRACE_GETREGS, p->pid, (PTRACE_ADDR_CAST)&reg, 0) < 0)
#endif
	{
		switch (errno)
		{
			case ESRCH:
				ret = PLUMP_ERROR_NOSUCHPID;
				break;
			default:
				ret = PLUMP_ERROR_FAILURE;
				break;
		}
	}
	else
	{
		plump_registersTranslate(REGISTERS_TYPE, (void *)&reg, registers, numRegisters, 1);

		ret = PLUMP_SUCCESS;
	}

	return ret;
}

#endif
